-----------------------------------------------------------------------
--
--  File:        svga.adb
--  Description: 256 color SVGA driver for VESA version 1.2 and higher
--  Rev:         0.1
--  Date:        03/09/96
--  Author:      Jerry van Dijk
--  Mail:        jerry@jvdsys.nextjk.stuyts.nl
--
--  Copyright (c) Jerry van Dijk, 1996
--  Forelstraat 211
--  2037 KV  HAARLEM
--  THE NETHERLANDS
--  tel int + 31 23 540 1052
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------

with MS_DOS; use MS_DOS;
with VESA_12; use VESA_12;
with Interfaces; use Interfaces;

package body SVGA is

   --------------------------------
   -- Define usuable video modes --
   --------------------------------
   subtype Video_Modes is SVGA_Mode range M640x400 .. M1280x1024;


   ----------------------------
   -- Define the system font --
   ----------------------------
   Font_Max_Char   : constant Natural := 255;
   Font_Max_Height : constant Natural := Font_Height - 1;

   type Font_Char is array (0 .. Font_Max_Height) of Unsigned_8;
   type Font_Type is array (0 .. Font_Max_Char) of Font_Char;


   ----------------------
   -- Global variables --
   ----------------------

   Bank_Step : Unsigned_16;
   Win_Base  : Unsigned_32;
   Line_Size : Unsigned_32;

   Package_Initialized : Boolean := False;
   In_Graphics_Mode    : Boolean := False;

   System_Font     : Font_Type;
   Startup_Mode    : Unsigned_16;
   Default_Palette : Color_Palette;
   Current_Bank    : Unsigned_16 := 0;
   Available_Modes : array (Video_Modes) of Boolean;


   ----------------------
   -- Global constants --
   ----------------------

   VESA_Signature : constant Unsigned_32 := 16#41534556#;


   -------------------------
   -- Intrinsic functions --
   -------------------------

   function Shift_Left (Value : Unsigned_8;
                        Count : Natural) return Unsigned_8;
   pragma Import (Intrinsic, Shift_Left);

   function Shift_Left (Value : Unsigned_16;
                        Count : Natural) return Unsigned_16;
   pragma Import (Intrinsic, Shift_Left);

   function Shift_Right (Value : Unsigned_32;
                         Count : Natural) return Unsigned_32;
   pragma Import (Intrinsic, Shift_Right);


   ------------------------------------------
   -- Check if switched to a graphics mode --
   ------------------------------------------

   procedure Check_In_Graphics_Mode is
   begin
      if not In_Graphics_Mode then
         raise Not_In_Graphics_Mode;
      end if;
   end Check_In_Graphics_Mode;
   pragma Inline (Check_In_Graphics_Mode);


   -------------------------------------------------------
   -- Check if screen coordinate are currently in range --
   -------------------------------------------------------

   procedure Check_In_Range (X : X_Loc; Y : Y_Loc) is
   begin
      if X > X_Max or Y > Y_Max then
         raise Out_Of_Range;
      end if;
   end Check_In_Range;
   pragma Inline (Check_In_Range);


   ------------------------------------------------
   -- Check if video mode is valid for this card --
   ------------------------------------------------

   function Is_Mode_Valid (Mode : Unsigned_16) return Boolean is
      Current_Mode : Unsigned_16;
      Mode_Table   : Unsigned_32;
      Vga_Info     : VgaInfoBlock;
      Mode_Info    : ModeInfoBlock;
      Result       : Boolean := False;
   begin
      Vga_Info.VESASignature := 16#41414141#;
      GetVgaInfoBlock (Vga_Info);
      Mode_Table := RM_To_Linear (Vga_Info.VideoModePtr);
      Set_Selector (DOS_Ds);
      Current_Mode := Peek_Word (Mode_Table);
      while Current_Mode /= 16#FFFF# and Current_Mode /= Mode loop
         Mode_Table := Mode_Table + 2;
         Current_Mode := Peek_Word (Mode_Table);
      end loop;
      if Current_Mode = Mode then
         GetModeInfoBlock (Mode, Mode_Info);
         if (Mode_Info.ModeAttributes and 16#01#) /= 0 then
            Result := True;
         end if;
      end if;
      return Result;
   end Is_Mode_Valid;


   --------------------------------------
   -- Copy system font from ROM to RAM --
   --------------------------------------

   procedure Load_System_Font is
      Regs    : Dpmi_Regs;
      Address : Unsigned_32;
   begin
      Regs.Ax := 16#1130#;
      Regs.Bx := 16#0200#;
      Dpmi_Int (16#10#, Regs);
      Address := 16 * Unsigned_32 (Regs.Es) + Unsigned_32 (Regs.Bp);
      for Num in 0 .. Font_Max_Char loop
         for Byte in 0 .. Font_Max_Height loop
            System_Font (Num)(Byte) := Peek_Byte (Address);
            Address := Address + 1;
         end loop;
      end loop;
   end Load_System_Font;


   --------------------------------
   -- Initialize the SVGA driver --
   --------------------------------

   procedure Init_Package is
      Driver     : VgaInfoBlock;
      Modes_Ok   : Boolean := False;
   begin
      Startup_Mode := GetCurrentVideoMode;
      Driver.VESASignature := 16#41414141#;
      GetVgaInfoBlock (Driver);
      if Driver.VESASignature /= VESA_Signature then
         raise No_VESA_Driver;
      end if;
      if Driver.VESAVersion /= 16#0102# and Driver.VESAVersion /= 16#0200# then
         raise Invalid_VESA_Version;
      end if;
      for Mode in Video_Modes'Range loop
         Available_Modes (Mode) := Is_Mode_Valid (SVGA_Mode'Enum_Rep (Mode));
         if Available_Modes (Mode) then
            Modes_Ok := True;
         end if;
      end loop;
      if not Modes_Ok then
         raise No_Valid_Modes;
      end if;
      Default_Palette := Get_Palette;
      Load_System_Font;
      Package_Initialized := True;
   end Init_Package;


   -------------------------------
   -- Put a pixel on the screen --
   -------------------------------

   procedure Private_Put_Pixel (X : X_Loc; Y : Y_Loc; Color : Color_Type) is
      Address : Unsigned_32 := Y * Line_Size + X;
      Bank : Unsigned_16 :=
        Bank_Step * Unsigned_16 (Shift_Right (Address, 16));
      Offset : Unsigned_32 := Win_Base + (Address and 16#FFFF#);
   begin
      if (X <= X_Max) and (Y <= Y_Max) then
         if Bank /= Current_Bank then
            Current_Bank := Bank;
            SetVideoMemoryWindow (0, Bank);
         end if;
         Poke_Byte (Offset, Color);
      end if;
   end Private_Put_Pixel;
   pragma Inline (Private_Put_Pixel);


   -----------------------------
   -- Bankswitch video memory --
   -----------------------------

   procedure Set_Bank (Bank : Unsigned_16) is
   begin
      if Bank /= Current_Bank then
         Current_Bank := Bank;
         SetVideoMemoryWindow (0, Bank);
      end if;
   end Set_Bank;
   pragma Inline (Set_Bank);


   -------------------------------------
   -- Calculate pixel bank and offset --
   -------------------------------------

   procedure Get_Address (X      : in     X_Loc;
                          Y      : in     Y_Loc;
                          Bank   :    out Unsigned_16;
                          Offset :    out Unsigned_32) is
      Address : Unsigned_32 := Line_Size * Y + X;
   begin
      Offset := Win_Base + (Address and 16#FFFF#);
      Bank := Bank_Step * Unsigned_16 (Shift_Right (Address, 16));
   end Get_Address;
   pragma Inline (Get_Address);


   ----------------------------------
   -- Set a new SVGA graphics mode --
   ----------------------------------

   procedure Graphics_Mode (Mode : SVGA_Mode := Autodetect) is
      Current_Mode : ModeInfoBlock;
      New_Mode     : SVGA_Mode := Mode;
   begin
      if not Package_Initialized then
         Init_Package;
      end if;
      if New_Mode = Autodetect then
         for I in reverse Video_Modes'Range loop
            if Available_Modes (I) then
               New_Mode := I;
               exit;
            end if;
         end loop;
      else
         if not Available_Modes (New_Mode) then
            raise Unsupported_Video_Mode;
         end if;
      end if;
      GetModeInfoBlock (SVGA_Mode'Enum_Rep (New_Mode), Current_Mode);
      X_Size := Natural (Current_Mode.XResolution);
      Y_Size := Natural (Current_Mode.YResolution);
      X_Max := Unsigned_32 (X_Size - 1);
      Y_Max := Unsigned_32 (Y_Size - 1);
      Win_Base := 16 * Unsigned_32 (Current_Mode.WinASegment);
      Line_Size := Unsigned_32 (Current_Mode.BytesPerScanLine);
      Bank_Step := Current_Mode.WinSize / Current_Mode.WinGranularity;
      SetSVGAMode (SVGA_Mode'Enum_Rep (New_Mode));
      In_Graphics_Mode := True;
   end Graphics_Mode;


   -------------------------------------
   -- Return to the -saved- text mode --
   -------------------------------------

   procedure Text_Mode is
   begin
      Check_In_Graphics_Mode;
      Restore_Default_Palette;
      SetSVGAMode (Startup_Mode);
      In_Graphics_Mode := False;
   end Text_Mode;


   --------------------------------------
   -- Set the color of a display pixel --
   --------------------------------------

   procedure Set_Pixel (X : X_Loc; Y : Y_Loc; Color : Color_Type) is
   begin
      Check_In_Graphics_Mode;
      Check_In_Range (X, Y);
      Set_Selector (DOS_Ds);
      Private_Put_Pixel (X, Y, Color);
   end Set_Pixel;


   ------------------------------
   -- Get the color of a Pixel --
   ------------------------------

   function Get_Pixel (X : X_Loc; Y : Y_Loc) return Color_Type is
      Value  : Color_Type;
      Bank   : Unsigned_16;
      Offset : Unsigned_32;
   begin
      Check_In_Graphics_Mode;
      Check_In_Range (X, Y);
      Set_Selector (DOS_Ds);
      Get_Address (X, Y, Bank, Offset);
      Set_Bank (Bank);
      Value := Peek_Byte (Offset);
      return Value;
   end Get_Pixel;

   -----------------------------------
   -- Get the current color palette --
   -----------------------------------

   function Get_Palette return Color_Palette is
      Palette : Color_Palette;
   begin
      Outport_Byte (16#3C7#, 0);
      for I in Color_Palette'Range loop
         Palette (I).Red := Inport_Byte (16#3C9#);
         Palette (I).Green := Inport_Byte (16#3C9#);
         Palette (I).Blue := Inport_Byte (16#3C9#);
      end loop;
      return Palette;
   end Get_Palette;


   -----------------------------
   -- Set a new color palette --
   -----------------------------

   procedure Set_Palette (Palette : Color_Palette) is
   begin
      Outport_Byte (16#3C8#, 0);
      for I in Color_Palette'Range loop
         Outport_Byte (16#3C9#, Palette (I).Red);
         Outport_Byte (16#3C9#, Palette (I).Green);
         Outport_Byte (16#3C9#, Palette (I).Blue);
      end loop;
   end Set_Palette;


   ---------------------------------
   -- Restore the default palette --
   ---------------------------------
   procedure Restore_Default_Palette is
   begin
      Set_Palette (Default_Palette);
   end Restore_Default_Palette;


   -----------------
   -- Set a color --
   -----------------

   procedure Set_Color (Color : Color_Type; Value : RGB_Color) is
   begin
      Outport_Byte (16#3C8#, Color);
      Outport_Byte (16#3C9#, Value.Red);
      Outport_Byte (16#3C9#, Value.Green);
      Outport_Byte (16#3C9#, Value.Blue);
   end Set_Color;


   -----------------
   -- Get a color --
   -----------------
   function Get_Color (Color : Color_Type) return RGB_Color is
      Value : RGB_Color;
   begin
      Outport_Byte (16#3C7#, Color);
      Value.Red := Inport_Byte (16#3C9#);
      Value.Green := Inport_Byte (16#3C9#);
      Value.Blue := Inport_Byte (16#3C9#);
      return Value;
   end Get_Color;


   ----------------------------
   -- Draw a horizontal line --
   ----------------------------

   procedure Hor_Line (X1, X2 : X_Loc; Y : Y_Loc; Color : Color_Type) is
      Left, Right : X_Loc;
   begin
      Check_In_Graphics_Mode;
      if X1 = X2 then
         Set_Pixel (X1, Y, Color);
      else
         Check_In_Range (X1, Y);
         Check_In_Range (X2, Y);
         Left := X1; Right := X2;
         if X1 > X2 then
            Left := X2; Right := X1;
         end if;
         Set_Selector (DOS_Ds);
         for X in Left .. Right loop
            Private_Put_Pixel (X, Y, Color);
         end loop;
      end if;
   end Hor_Line;


   --------------------------
   -- Draw a vertical line --
   --------------------------

   procedure Ver_Line (X : X_Loc; Y1, Y2 : Y_Loc; Color : Color_Type) is
      Top, Bottom : Y_Loc;
   begin
      Check_In_Graphics_Mode;
      if Y1 = Y2 then
         Set_Pixel (X, Y1, Color);
      else
         Check_In_Range (X, Y1);
         Check_In_Range (X, Y2);
         Top := Y1; Bottom := Y2;
         if Y1 > Y2 then
            Top := Y2; Bottom := Y1;
         end if;
         Set_Selector (DOS_Ds);
         for Y in Top .. Bottom loop
            Private_Put_Pixel (X, Y, Color);
         end loop;
      end if;
   end Ver_Line;


   -----------------
   -- Draw a line --
   -----------------

   procedure Line (X1 : X_Loc; Y1 : Y_Loc;
                   X2 : X_Loc; Y2 : Y_Loc; Color : Color_Type) is
      Dir    : Boolean := False;
      Left   : Integer := Integer (X1);
      Right  : Integer := Integer (X2);
      Top    : Integer := Integer (Y1);
      Bottom : Integer := Integer (Y2);
      Temp, Limit, D_X, D_Y, S_X, S_Y : Integer;
   begin
      Check_In_Graphics_Mode;
      if X1 = X2 and Y1 = Y2 then
         Set_Pixel (X1, Y1, Color);
      elsif X1 = X2 then
         Ver_Line (X1, Y1, Y2, Color);
      elsif Y1 = Y2 then
         Hor_Line (X1, X2, Y1, Color);
      else
         Check_In_Range (X1, Y1);
         Check_In_Range (X2, Y2);
         D_X := abs (Right - Left);
         S_X := 1;
         if Left > Right then
            S_X := -1;
         end if;
         D_Y := abs (Bottom - Top);
         S_Y := 1;
         if Top > Bottom then
            S_Y := -1;
         end if;
         if D_Y > D_X then
            Dir := True;
            Temp := Left;
            Left := Right;
            Right := Temp;
            Temp := D_X;
            D_X := D_Y;
            D_Y := Temp;
            Temp := S_X;
            S_X := S_Y;
            S_Y := Temp;
         end if;
         Limit := 2 * D_Y - D_X;
         for I in 0 .. D_X loop
            if Dir then
               Private_Put_Pixel (Unsigned_32 (Top),
                                  Unsigned_32 (Left), Color);
            else
               Private_Put_Pixel (Unsigned_32 (Left),
                                  Unsigned_32 (Top), Color);
            end if;
            while Limit >= 0 loop
               Top := Top + S_Y;
               Limit := Limit - 2 * D_X;
            end loop;
            Left := Left + S_X;
            Limit := Limit + 2 * D_Y;
         end loop;
         Private_Put_Pixel (Unsigned_32 (Right), Unsigned_32 (Bottom), Color);
      end if;
   end Line;


   ----------------------
   -- Clear the screen --
   ----------------------

   procedure Clear_Screen (Color : Color_Type := Standard_Black) is
      Bank        : Unsigned_16;
      Offset      : Unsigned_32;
      Loop_High   : Unsigned_32 := Win_Base + 16#FFFF#;
      Screen_Size : Unsigned_32 := Unsigned_32 (Y_Size) * Line_Size;
      Full_Banks  : Unsigned_32 := Shift_Right (Screen_Size, 16);
      Bytes_Left  : Unsigned_32 := Win_Base + (Screen_Size and 16#FFFF#);
   begin
      Check_In_Graphics_Mode;
      Set_Selector (DOS_Ds);
      Get_Address (0, 0, Bank, Offset);
      Set_Bank (Bank);
      for I in 0 .. Full_Banks loop
         for J in Win_Base .. Loop_High loop
            Poke_Byte (Unsigned_32 (J), Color);
         end loop;
         Bank := Bank + Bank_Step;
         Set_Bank (Bank);
      end loop;
      for J in Win_Base .. Bytes_Left loop
         Poke_Byte (Unsigned_32 (J), Color);
      end loop;
   end Clear_Screen;


   ----------------------
   -- Fill a rectangle --
   ----------------------

   procedure Fill_Rect (X1 : X_Loc; Y1 : Y_Loc;
                        X2 : X_Loc; Y2 : Y_Loc; Color : Color_Type) is
      Left, Right : X_Loc;
      Top, Bottom : Y_Loc;
   begin
      Check_In_Graphics_Mode;
      if X1 = X2 and Y1 = Y2 then
         Set_Pixel (X1, Y1, Color);
      elsif X1 = X2 then
         Ver_Line (X1, Y1, Y2, Color);
      elsif Y1 = Y2 then
         Hor_Line (X1, X2, Y1, Color);
      else
         Check_In_Range (X1, Y1);
         Check_In_Range (X2, Y2);
         Left := X1; Right := X2;
         if X1 > X2 then
            Left := X2; Right := X1;
         end if;
         Top := Y1; Bottom := Y2;
         if Y1 > Y2 then
            Top := Y2; Bottom := Y1;
         end if;
         Set_Selector (DOS_Ds);
         for Y in Top .. Bottom loop
            for X in Left .. Right loop
               Private_Put_Pixel (X, Y, Color);
            end loop;
         end loop;
      end if;
   end Fill_Rect;


   ------------------------------------
   -- Draw a visually correct circle --
   ------------------------------------

   procedure Circle (X : X_Loc; Y : Y_Loc;
                     Radius : Positive; Color : Color_Type) is
      C1, C2  : Unsigned_32;
      Count   : Integer := 0;
      Current : Integer := Radius;
      Diam    : Integer := 2 * (1 - Radius);
      Correct : Integer := 2 * X_Size / Y_Size;
   begin
      Check_In_Graphics_Mode;
      Set_Selector (DOS_Ds);
      if Radius = 1 then
         Set_Pixel (X, Y, Color);
      else
         Check_In_Range (X, Y);
         while Current >= 0 loop
            C1 := Unsigned_32 (Count);
            C2 := Unsigned_32 (Current);
            Private_Put_Pixel (X + C1, Y + C2, Color);
            Private_Put_Pixel (X + C1, Y - C2, Color);
            Private_Put_Pixel (X - C1, Y + C2, Color);
            Private_Put_Pixel (X - C1, Y - C2, Color);
            if (Diam + Current) > 0 then
               Current := Current - 1;
               Diam := Diam - (Correct * Current) - 1;
            end if;
            if (Count > Diam) then
               Count := Count + 1;
               Diam := Diam + (2 * Count) + 1;
            end if;
         end loop;
      end if;
   end Circle;


   ------------------------------------------
   -- Draw a Mathematically correct circle --
   ------------------------------------------

   procedure Math_Circle (X : X_Loc; Y : Y_Loc;
                          Radius : Positive; Color : Color_Type) is
      C1, C2 : Unsigned_32;
      X1     : Integer := 0;
      Y1     : Integer := Radius;
      Diff   : Integer := 2 * (1 - Radius);
   begin
      Check_In_Graphics_Mode;
      Set_Selector (DOS_Ds);
      if Radius = 1 then
         Set_Pixel (X, Y, Color);
      else
         Check_In_Range (X, Y);
         Private_Put_Pixel (X + Unsigned_32 (Radius), Y, Color);
         Private_Put_Pixel (X - Unsigned_32 (Radius), Y, Color);
         while Y1 > 0 loop
            C1 := Unsigned_32 (X1);
            C2 := Unsigned_32 (Y1);
            Private_Put_Pixel (X + C1, Y + C2, Color);
            Private_Put_Pixel (X + C1, Y - C2, Color);
            Private_Put_Pixel (X - C1, Y + C2, Color);
            Private_Put_Pixel (X - C1, Y - C2, Color);
            if Diff + Y1 > 0 then
               Y1 := Y1 - 1;
               Diff := Diff - 2 * Y1 - 1;
            end if;
            if X1 > Diff then
               X1 := X1 + 1;
               Diff := Diff + 2 * X1 + 1;
            end if;
         end loop;
      end if;
   end Math_Circle;


   -----------------------------------
   -- Write a string to the display --
   -----------------------------------

   procedure Write_String (X : X_Loc; Y : Y_Loc;
                           S : String; Color : Color_Type := Standard_White) is
      Current : Unsigned_8;
   begin
      Check_In_Graphics_Mode;
      Check_In_Range (X, Y);
      Check_In_Range (X + Unsigned_32 (S'Length * Font_Width) - 1,
                      Y + Unsigned_32 (Font_Max_Height));
      Set_Selector (DOS_Ds);
      for Count in S'First .. S'Last loop
         for Line in Font_Char'Range loop
            Current := System_Font (Character'Pos (S (Count)))(Line);
            for Bit in 0 .. Font_Width - 1 loop
               if (Current and 16#80#) /= 0 then
                  Private_Put_Pixel
                    (X + Unsigned_32 (Count * Font_Width + Bit),
                     Y + Unsigned_32 (Line), Color);
               end if;
               Current := Shift_Left (Current, 1);
            end loop;
         end loop;
      end loop;
   end Write_String;


   --------------------------------
   -- Initialize a screen buffer --
   --------------------------------

   procedure Initialize (SB : in out Screen_Buffer) is
   begin
      if SB.Data /= null then
         raise Storage_Error;
      end if;
      if SB.Width = 0 or SB.Height = 0 or
         SB.Width > 1279 or SB.Height > 1023 then
         raise Out_Of_Range;
      end if;
      SB.Data := new Data_Buffer (0 .. SB.Width * SB.Height - 1);
   end Initialize;


   ------------------------------
   -- Finalize a screen buffer --
   ------------------------------

   procedure Finalize (SB : in out Screen_Buffer) is
   begin
      Free (SB.Data);
      SB.Data := null;
   end Finalize;


   ---------------------------------------------
   -- Load a screen area into a screen buffer --
   ---------------------------------------------

   procedure Get_Buffer (X : X_Loc;
                         Y : Y_Loc; Buffer : in out Screen_Buffer) is
      Bank   : Unsigned_16;
      Offset : Unsigned_32;
      Index  : Natural := 0;
      X2     : X_Loc := X + X_Loc (Buffer.Width - 1);
      Y2     : Y_Loc := Y + Y_Loc (Buffer.Height - 1);
   begin
      Check_In_Graphics_Mode;
      if Buffer.Data = null then
         raise Storage_Error;
      end if;
      Check_In_Range (X, Y);
      Check_In_Range (X2, Y2);
      Set_Selector (DOS_Ds);
      for I in Y .. Y2 loop
         for J in X .. X2 loop
            Get_Address (I, J, Bank, Offset);
            Set_Bank (Bank);
            Buffer.Data (Index) := Peek_Byte (Offset);
            Index := Index + 1;
         end loop;
      end loop;
   end Get_Buffer;


   -----------------------------------
   -- Put a Screen_Buffer on screen --
   -----------------------------------

   procedure Put_Buffer (X : X_Loc; Y : Y_Loc; Buffer : Screen_Buffer) is
      Bank   : Unsigned_16;
      Offset : Unsigned_32;
      Index  : Natural := 0;
      X2     : X_Loc := X + X_Loc (Buffer.Width - 1);
      Y2     : Y_Loc := Y + Y_Loc (Buffer.Height - 1);
   begin
      Check_In_Graphics_Mode;
      if Buffer.Data = null then
         raise Storage_Error;
      end if;
      Check_In_Range (X, Y);
      Check_In_Range (X2, Y2);
      Set_Selector (DOS_Ds);
      for I in Y .. Y2 loop
         for J in X .. X2 loop
            Get_Address (I, J, Bank, Offset);
            Set_Bank (Bank);
            Poke_Byte (Offset, Buffer.Data (Index));
            Index := Index + 1;
         end loop;
      end loop;
   end Put_Buffer;

end SVGA;