-----------------------------------------------------------------------
--
--  File:        svga.ads
--  Description: 256 color SVGA driver for VESA version 1.2 and higher
--  Rev:         0.1
--  Date:        03/09/96
--  Author:      Jerry van Dijk
--  Mail:        jerry@jvdsys.nextjk.stuyts.nl
--
--  Copyright (c) Jerry van Dijk, 1996
--  Forelstraat 211
--  2037 KV  HAARLEM
--  THE NETHERLANDS
--  tel int + 31 23 540 1052
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------

with Ada.Unchecked_Deallocation, Ada.Finalization, Interfaces;

package SVGA is

   ---------------------------------------
   -- Specify the supported video modes --
   ---------------------------------------

   type SVGA_Mode is (M640x400, M640x480, M800x600,
                      M1024x768, M1280x1024, Autodetect);
   for SVGA_Mode use (M640x400 => 16#0100#, M640x480 => 16#0101#,
                      M800x600 => 16#0103#, M1024x768 => 16#0105#,
                      M1280x1024 => 16#0107#, Autodetect => 16#FFFF#);
   for SVGA_Mode'Size use 16;


   ---------------------------
   -- Specify color palette --
   ---------------------------

   Num_Colors : constant := 256;

   subtype Color_Type is Interfaces.Unsigned_8;
   subtype Color_Value is Interfaces.Unsigned_8 range 0 .. 63;

   type RGB_Color is
      record
         Red   : Color_Value;
         Green : Color_Value;
         Blue  : Color_Value;
      end record;

   type Color_Palette is array (Color_Type) of RGB_Color;

   --  The standard colors below are only valid for the default palette!

   Standard_Black         : constant Color_Type := 0;
   Standard_Blue          : constant Color_Type := 1;
   Standard_Green         : constant Color_Type := 2;
   Standard_Cyan          : constant Color_Type := 3;
   Standard_Red           : constant Color_Type := 4;
   Standard_Magenta       : constant Color_Type := 5;
   Standard_Brown         : constant Color_Type := 6;
   Standard_Light_Gray    : constant Color_Type := 7;
   Standard_Dark_Gray     : constant Color_Type := 8;
   Standard_Light_Blue    : constant Color_Type := 9;
   Standard_Light_Green   : constant Color_Type := 10;
   Standard_Light_Cyan    : constant Color_Type := 11;
   Standard_Light_Red     : constant Color_Type := 12;
   Standard_Light_Magenta : constant Color_Type := 13;
   Standard_Yellow        : constant Color_Type := 14;
   Standard_White         : constant Color_Type := 15;


   --------------------------
   -- Specify display size --
   --------------------------

   subtype Y_Loc is Interfaces.Unsigned_32 range 0 .. 1023;
   subtype X_Loc is Interfaces.Unsigned_32 range 0 .. 1279;


   ---------------------------------------
   -- Specify actual display parameters --
   ---------------------------------------

   X_Max : X_Loc;
   Y_Max : Y_Loc;
   X_Size, Y_Size : Natural;


   ------------------------------
   -- Specify system font size --
   ------------------------------

   Font_Width  : constant Integer :=  8;
   Font_Height : constant Integer := 14;


   -----------------------------
   -- Specify a screen buffer --
   -----------------------------

   type Screen_Buffer (New_Width  : Positive;
                       New_Height : Positive) is limited private;


   -----------------------------------
   -- SVGA functions and procedures --
   -----------------------------------

   --  Display functions
   procedure Graphics_Mode (Mode : SVGA_Mode := Autodetect);
   procedure Text_Mode;

   --  Palette functions
   procedure Restore_Default_Palette;
   function Get_Palette return Color_Palette;
   procedure Set_Palette (Palette : Color_Palette);
   function Get_Color (Color : Color_Type) return RGB_Color;
   procedure Set_Color (Color : Color_Type; Value : RGB_Color);

   --  Pixel functions
   function Get_Pixel (X : X_Loc; Y : Y_Loc) return Color_Type;
   procedure Set_Pixel (X : X_Loc; Y : Y_Loc; Color : Color_Type);

   --  Line functions
   procedure Hor_Line (X1, X2 : X_Loc; Y : Y_Loc; Color : Color_Type);
   procedure Ver_Line (X : X_Loc; Y1, Y2 : Y_Loc; Color : Color_Type);
   procedure Line (X1 : X_Loc; Y1 : Y_Loc;
                   X2 : X_Loc; Y2 : Y_Loc; Color : Color_Type);

   --  Circle functions
   procedure Circle (X : X_Loc; Y : Y_Loc;
                     Radius : Positive; Color : Color_Type);
   procedure Math_Circle (X : X_Loc; Y : Y_Loc;
                          Radius : Positive; Color : Color_Type);

   --  Area functions
   procedure Clear_Screen (Color : Color_Type := Standard_Black);
   procedure Fill_Rect (X1 : X_Loc; Y1 : Y_Loc;
                        X2 : X_Loc; Y2 : Y_Loc; Color : Color_Type);


   --  Text functions
   procedure Write_String (X : X_Loc; Y : Y_Loc;
                           S : String; Color : Color_Type := Standard_White);

   --  Screen buffer functions
   procedure Put_Buffer (X : X_Loc; Y : Y_Loc; Buffer : Screen_Buffer);
   procedure Get_Buffer (X : X_Loc; Y : Y_Loc; Buffer : in out Screen_Buffer);


   ---------------------------------
   -- Package specific exceptions --
   ---------------------------------

   Out_Of_Range           : exception;
   No_VESA_Driver         : exception;
   No_Valid_Modes         : exception;
   Invalid_VESA_Version   : exception;
   Not_In_Graphics_Mode   : exception;
   Unsupported_Video_Mode : exception;

private

   -----------------------
   -- The screen buffer --
   -----------------------

   type Data_Buffer is array (Natural range <>) of Interfaces.Unsigned_8;
   type Data_Buffer_Access is access all Data_Buffer;

   procedure Free is
     new Ada.Unchecked_Deallocation (Data_Buffer, Data_Buffer_Access);

   type Screen_Buffer (New_Width  : Positive;
                       New_Height : Positive) is
     new Ada.Finalization.Limited_Controlled with
      record
         Width  : Positive := New_Width;
         Height : Positive := New_Height;
         Data   : Data_Buffer_Access;
      end record;

   procedure Finalize (SB : in out Screen_Buffer);
   procedure Initialize (SB : in out Screen_Buffer);


   --------------------------------------
   -- Private procedures and functions --
   --------------------------------------

   procedure Init_Package;
   procedure Load_System_Font;
   procedure Check_In_Graphics_Mode;
   procedure Check_In_Range (X : X_Loc; Y : Y_Loc);
   procedure Set_Bank (Bank : Interfaces.Unsigned_16);
   function Is_Mode_Valid (Mode : Interfaces.Unsigned_16) return Boolean;
   procedure Private_Put_Pixel (X : X_Loc; Y : Y_Loc; Color : Color_Type);

   procedure Get_Address (X      : in     X_Loc;
                          Y      : in     Y_Loc;
                          Bank   :    out Interfaces.Unsigned_16;
                          Offset :    out Interfaces.Unsigned_32);

end SVGA;