-----------------------------------------------------------------------
--
--  File:        vesa_12.adb
--  Description: Interface to VESA v1.2 driver
--  Rev:         0.1
--  Date:        03/09/96
--  Author:      Jerry van Dijk
--  Mail:        jerry@jvdsys.nextjk.stuyts.nl
--
--  Copyright (c) Jerry van Dijk, 1996
--  Forelstraat 211
--  2037 KV  HAARLEM
--  THE NETHERLANDS
--  tel int + 31 23 540 1052
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------

with MS_DOS; use MS_DOS;

package body VESA_12 is

   ---------------------------------------
   -- Maximum size of a VESA info block --
   ---------------------------------------

   --  VESA v1.2 always returns a 256-byte block, but a otherwise
   --  compatible v2.0 implementation /might/ return 512 bytes

   Max_Infoblock_Length : constant := 512;


   ----------------------------
   -- VESA related constants --
   ----------------------------

   VESA_Call_Ok : constant Unsigned_16 := 16#004F#;
   --  When Ax has this value, the VESA call succeeded

   Video_Interrupt : constant Unsigned_16 := 16#0010#;
   --  The BIOS video interupt, used to call the VESA driver


   -------------------------------------
   -- Get the VESA driver information --
   -------------------------------------

   --  Pre-fill the VESASignature field of the VgaInfoBlock with "AAAA"
   --  to make sure that later versions return a v1.2 compatible
   --  infoblock

   procedure GetVgaInfoBlock (Info : in out VgaInfoBlock) is
      Regs : Dpmi_Regs;
   begin
      --  Make sure the size of the real-mode transfer buffer is
      --  large enough to hold the VgaInfoBlock
      if TB_Size < Max_Infoblock_Length then
         raise Storage_Error;
      end if;

      --  Load the VgaInfoBlock into the real-mode transfer buffer
      Regs.Ax := 16#4F00#;
      Regs.Es := RM_Segment (DOS_Extender.Linear_Address_Of_Transfer_Buffer);
      Regs.Di := RM_Offset (DOS_Extender.Linear_Address_Of_Transfer_Buffer);
      Dpmi_Int (Video_Interrupt, Regs);

      --  Check if VESA call succeeded
      if Regs.Ax /= VESA_Call_Ok then
         raise VESA_Error;
      end if;

      --  Copy the VgaInfoBlock in real-mode transfer buffer into the
      --  VgaInfoBlock variable
      Dosmemget (TB_Address, Info'Size / 8, To_Unsigned_32 (Info'Address));
   end GetVgaInfoBlock;


   ---------------------------------------
   -- Get the Mode info on a video mode --
   ---------------------------------------

   --  The Mode must be one that is present in the video mode table
   --  pointed to by the VgaInfoBlock VideoModePtr

   procedure GetModeInfoBlock (Mode : Unsigned_16; Info : out ModeInfoBlock) is
      Regs : Dpmi_Regs;
   begin
      --  Make sure the size of the real-mode transfer buffer is
      --  large enough to hold the ModeInfoBlock
      if TB_Size < Max_Infoblock_Length then
         raise Storage_Error;
      end if;

      --  Load the ModeInfoBlock for this mode into the real-mode
      --  transfer buffer
      Regs.Ax := 16#4F01#;
      Regs.Cx := Mode;
      Regs.Es := RM_Segment (TB_Address);
      Regs.Di := RM_Offset (TB_Address);
      Dpmi_Int (Video_Interrupt, Regs);

      --  Check if VESA call succeeded
      if Regs.Ax /= VESA_Call_Ok then
         raise VESA_Error;
      end if;

      --  Copy the ModeInfoBlock in real-mode transfer buffer into the
      --  ModeInfoBlock variable
      Dosmemget (TB_Address, Info'Size / 8, To_Unsigned_32 (Info'Address));
   end GetModeInfoBlock;


   --------------------------
   -- Set a new SVGA Mode --
   --------------------------

   --  The Mode must be either on  that is present in the video mode
   --  table pointed to by the VgaInfoBlock VideoModePtr or a standard
   --  CGA/EGA/VGA mode.

   --  Note that the mode is 16-bit wide and the clear flag is bit 15

   procedure SetSVGAMode (Mode : Unsigned_16) is
      Regs : Dpmi_Regs;
   begin
      --  Ask the VESA driver to switch the display mode
      Regs.Ax := 16#4F02#;
      Regs.Bx := Mode;
      Dpmi_Int (Video_Interrupt, Regs);

      --  Check if VESA call succeeded
      if Regs.Ax /= VESA_Call_Ok then
         raise VESA_Error;
      end if;
   end SetSVGAMode;


   ----------------------------
   -- Get current video mode --
   ----------------------------

   --  Does not return the clear screen status!

   function GetCurrentVideoMode return Unsigned_16 is
      Regs : Dpmi_Regs;
   begin
      --  Ask the VESA driver to figure out the video mode
      Regs.Ax := 16#4F03#;
      Dpmi_Int (Video_Interrupt, Regs);

      --  Check if VESA call succeeded
      if Regs.Ax /= VESA_Call_Ok then
         raise VESA_Error;
      end if;

      --  Return the video mode
      return Regs.Bx;
   end GetCurrentVideoMode;


   -------------------------------------
   -- Change the video memory mapping --
   -------------------------------------

   --  The Window is either 0 (window A) or 1 (window B)
   --  The new bank is specified as an offset expressed in
   --  Window Granularity units.

   procedure SetVideoMemoryWindow (Window : Unsigned_16; Pos : Unsigned_16) is
      Regs : Dpmi_Regs;
   begin
      --  Call the VESA driver to change the mapping
      Regs.Ax := 16#4F05#;
      Regs.Bx := Window;
      Regs.Dx := Pos;
      Dpmi_Int (Video_Interrupt, Regs);

      --  Check if VESA call succeeded
      if Regs.Ax /= VESA_Call_Ok then
         raise VESA_Error;
      end if;
   end SetVideoMemoryWindow;

end VESA_12;